@@ -21,22 +21,31 @@
 #include <config.h>
 #endif
 
+#include <math.h>
+
 #include <gdk/gdk.h>
 #include <gtk/gtkbutton.h>
 #include <gtk/gtkimage.h>
 
 #include "xfce_iconbutton.h"
 
+#define DEBUGGING 0
+
+/* XXX: this is used by gtkbutton. hope they won't change it soon ... */
+#define CHILD_SPACING 1
+
 /* Forward declarations */
 
 static void xfce_iconbutton_class_init (XfceIconbuttonClass * class);
 static void xfce_iconbutton_init (XfceIconbutton * iconbutton);
 static void xfce_iconbutton_finalize (GObject * object);
-static void xfce_iconbutton_size_allocate (GtkWidget * widget,
-				           GtkAllocation * allocation);
 
-static void xfce_iconbutton_scale (XfceIconbutton * iconbutton, int width,
-				   int height);
+static void xfce_iconbutton_size_allocate (GtkWidget * widget,
+					   GtkAllocation * allocation);
+
+static gboolean xfce_iconbutton_scale (XfceIconbutton * iconbutton);
+
+static void image_style_set (GtkWidget * widget, GtkStyle * old);
 
 /* Local data */
 
@@ -83,7 +92,7 @@ xfce_iconbutton_class_init (XfceIconbuttonClass * class)
     parent_class = gtk_type_class (gtk_button_get_type ());
 
     widget_class->size_allocate = xfce_iconbutton_size_allocate;
-    
+
     gobject_class->finalize = xfce_iconbutton_finalize;
 }
 
@@ -94,9 +103,13 @@ xfce_iconbutton_init (XfceIconbutton * iconbutton)
     gtk_widget_show (iconbutton->image);
     gtk_container_add (GTK_CONTAINER (iconbutton), iconbutton->image);
 
-    iconbutton->icon_size = -1;
+    iconbutton->icon_width = iconbutton->icon_height = -1;
     iconbutton->pb = NULL;
-    iconbutton->scaled = FALSE;
+
+    /* force style thickness programmatically */
+    g_signal_connect (iconbutton->image, "style-set",
+		      G_CALLBACK (image_style_set), NULL);
+
 }
 
 GtkWidget *
@@ -128,16 +141,14 @@ xfce_iconbutton_set_pixbuf (XfceIconbutton * iconbutton, GdkPixbuf * pixbuf)
     iconbutton->pb = pixbuf;
     g_object_ref (iconbutton->pb);
 
-    if (iconbutton->icon_size > 0)
+    if (iconbutton->icon_width > 0)
     {
-        xfce_iconbutton_scale (iconbutton,
-                               iconbutton->icon_size, 
-                               iconbutton->icon_size);
+	xfce_iconbutton_scale (iconbutton);
     }
     else
     {
-        gtk_image_set_from_pixbuf (GTK_IMAGE (iconbutton->image), 
-                                   iconbutton->pb);
+	gtk_image_set_from_pixbuf (GTK_IMAGE (iconbutton->image),
+				   iconbutton->pb);
     }
 }
 
@@ -160,114 +171,133 @@ xfce_iconbutton_finalize (GObject * object)
     G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
-static void 
+static void
 xfce_iconbutton_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
 {
     XfceIconbutton *iconbutton;
-    gint xthick, ythick;
-    GtkAllocation child_allocation;
+    gint xthick, ythick, borderwidth;
     GtkBorder *tmp_border;
 
     g_return_if_fail (widget != NULL);
     g_return_if_fail (XFCE_IS_ICONBUTTON (widget));
     g_return_if_fail (allocation != NULL);
 
-    iconbutton = XFCE_ICONBUTTON(widget);
-    
-    xthick = widget->style->xthickness;
-    ythick = widget->style->ythickness;
+    iconbutton = XFCE_ICONBUTTON (widget);
+
+    borderwidth = gtk_container_get_border_width (GTK_CONTAINER (widget));
+
+    xthick = 2 * (CHILD_SPACING + borderwidth + widget->style->xthickness +
+		  iconbutton->image->style->xthickness);
+    ythick = 2 * (CHILD_SPACING + borderwidth + widget->style->ythickness +
+		  iconbutton->image->style->ythickness);
 
     gtk_widget_style_get (widget, "default_border", &tmp_border, NULL);
 
-    if (tmp_border)
+    if (tmp_border && GTK_WIDGET_CAN_DEFAULT (widget))
     {
-	xthick += (tmp_border->left + tmp_border->right)/2;
-	ythick += (tmp_border->top + tmp_border->bottom)/2;
-	g_free(tmp_border);
+	xthick += tmp_border->left + tmp_border->right;
+	ythick += tmp_border->top + tmp_border->bottom;
+	gtk_widget_queue_draw(widget);
+    }
+
+    if (iconbutton->icon_width == allocation->width - xthick &&
+	iconbutton->icon_height == allocation->height - ythick)
+    {
+	parent_class->size_allocate (widget, allocation);
+	return;
     }
     
-    iconbutton->icon_size = MIN(allocation->width - 2 * xthick, 
-                                allocation->height - 2 * ythick);
+    iconbutton->icon_width = allocation->width - xthick;
+    iconbutton->icon_height = allocation->height - ythick;
 
     widget->allocation = *allocation;
 
-#if 0
-    g_message("size: %d x %d, scaled: %s", 
-	      allocation->width,
-	      allocation->height,
-	      iconbutton->scaled ? "yes" : "no");
+#if DEBUGGING
+    g_print (" ** size: %d x %d\n",
+	     iconbutton->icon_width,
+	     iconbutton->icon_height);
 #endif
-
-    /* prevent race condition */
-    if (!iconbutton->scaled)
-    {
-	xfce_iconbutton_scale (iconbutton, 
-			       iconbutton->icon_size,
-			       iconbutton->icon_size);
-	iconbutton->scaled = TRUE;
-    }
-    else
-    {
-	iconbutton->scaled = FALSE;
-    }
-
-    child_allocation = *allocation;
-    child_allocation.width = child_allocation.height = iconbutton->icon_size;
-    gtk_widget_size_allocate(iconbutton->image, &child_allocation);
+    xfce_iconbutton_scale (iconbutton);
 
     parent_class->size_allocate (widget, allocation);
-    
-#if 0
-    g_message("pixbuf size: %d", iconbutton->icon_size);
-    g_message("image size: %d x %d", 
-	      iconbutton->image->allocation.width,
-	      iconbutton->image->allocation.height);
+
+#if DEBUGGING
+    g_print (" ** allocated size: %d x %d\n",
+	     iconbutton->image->allocation.width,
+	     iconbutton->image->allocation.height);
 #endif
 }
 
-static void
-xfce_iconbutton_scale (XfceIconbutton * iconbutton, int width, int height)
+/* returns TRUE when image is set, FALSE otherwise */
+static gboolean
+xfce_iconbutton_scale (XfceIconbutton * iconbutton)
 {
     GdkPixbuf *tmp;
-    float wratio, hratio;
+    double wratio, hratio;
     int w, h, pw, ph;
 
-    g_return_if_fail (iconbutton != NULL);
-    g_return_if_fail (XFCE_IS_ICONBUTTON (iconbutton));
-    g_return_if_fail (iconbutton->pb != NULL);
-    g_return_if_fail (GDK_IS_PIXBUF (iconbutton->pb));
+    g_return_val_if_fail (iconbutton != NULL, FALSE);
+    g_return_val_if_fail (XFCE_IS_ICONBUTTON (iconbutton), FALSE);
+    g_return_val_if_fail (iconbutton->pb != NULL, FALSE);
+    g_return_val_if_fail (GDK_IS_PIXBUF (iconbutton->pb), FALSE);
 
     pw = gdk_pixbuf_get_width (iconbutton->pb);
     ph = gdk_pixbuf_get_height (iconbutton->pb);
 
-    w = width <= 0 ? pw + 1 : width;
-    h = height <= 0 ? ph + 1 : height;
+    w = iconbutton->icon_width <= 0 ? pw : iconbutton->icon_width;
+    h = iconbutton->icon_height <= 0 ? ph : iconbutton->icon_height;
 
-    wratio = (float) pw / (float) w;
-    hratio = (float) ph / (float) h;
-
-#if 0
-    g_message("w = %d, h = %d, pw = %d, ph = %d\n", w, h, pw, ph);
-    g_message("wratio = %f, hratio = %f\n", wratio, hratio);
+    if (w >= pw && h >= ph)
+    {
+#if DEBUGGING
+	g_print (" ** don't scale: %d x %d\n", pw, ph);
 #endif
+	/* don't scale up -> ugly icons */ 
+	tmp = iconbutton->pb;
+	g_object_ref (tmp);
+    }
+    else
+    {
+	wratio = (double) pw / w;
+	hratio = (double) ph / h;
 
-    if (wratio <= 1.0 && hratio <= 1.0)
-    {
-	/* don't scale up -> ugly icons */
-	gtk_image_set_from_pixbuf (GTK_IMAGE (iconbutton->image),
-				   iconbutton->pb);
-	return;
-    }
-    else if (wratio > hratio)
-    {
-	h = (float) ph / wratio;
-    }
-    else if (hratio > wratio)
-    {
-	w = (float) pw / hratio;
+	if (wratio > hratio)
+	    h = rint ((double) ph / wratio);
+	else if (hratio > wratio)
+	    w = rint ((double) pw / hratio);
+
+#if DEBUGGING
+	g_print (" ** scale image: %d x %d\n", w, h);
+#endif
+	tmp = gdk_pixbuf_scale_simple (iconbutton->pb, w, h,
+				       GDK_INTERP_BILINEAR);
     }
 
-    tmp = gdk_pixbuf_scale_simple (iconbutton->pb, w, h, GDK_INTERP_BILINEAR);
     gtk_image_set_from_pixbuf (GTK_IMAGE (iconbutton->image), tmp);
     g_object_unref (tmp);
+
+    return TRUE;
+}
+
+static void
+real_style_set (GtkWidget * widget, GtkStyle * old)
+{
+    GtkRcStyle *style;
+
+    style = gtk_widget_get_modifier_style (widget);
+    style->xthickness = style->ythickness = 0;
+    gtk_widget_modify_style (widget, style);
+}
+
+static void
+image_style_set (GtkWidget * widget, GtkStyle * old)
+{
+    static int recursive = 0;
+
+    if (recursive > 0)
+	return;
+
+    recursive++;
+    real_style_set (widget, old);
+    recursive--;
 }
